﻿# -*- coding: utf-8 -*-

from Npp import editor, editor1, editor2, notepad, NOTIFICATION, SCINTILLANOTIFICATION, INDICATORSTYLE
import ctypes
import ctypes.wintypes as wintypes

from collections import OrderedDict
regexes = OrderedDict()

# ------------------------------------------------- configuration area ---------------------------------------------------
#
# Define the lexer name exactly as it can be found in the Language menu
lexer_name = 'JRiver Media Center 2020 v2'

# Definition of colors and regular expressions
#   Note, the order in which regular expressions will be processed
#   is determined by its creation, that is, the first definition is processed first, then the 2nd, and so on
#
#   The basic structure always looks like this
#
#   regexes[(a, b)] = (c, d)
#
#   regexes = an ordered dictionary which ensures that the regular expressions are always processed in the same order
#   a = a unique number - suggestion, start with 0 and always increase by one
#   b = color in the form of (r,g,b) such as (255,0,0) for the color red
#   c = raw byte string, describes the regular expression. Example r'\w+'
#   d = number of the match group to be used


# Regex Definitions:
regexes[(0, (100, 100, 100))] = (r'(?i)(?<!\/)\[[^=\]]*\]', 0)		#default coloring for fields; anything recognized will be recolored later
#regexes[(0, (100, 100, 100))] = (r'(?i)\[\K[^=\]]*(?=\])', 0)	#default coloring for fields (uncolored brackets); anything recognized will be recolored later

regexes[(1, (255, 0, 255))] = (r'(?i)\[v_[^\]]*\]', 0)		#detects variables with field reference
regexes[(2, (255, 128, 0))] = (r'(?i)((?<=<font)([\s\S]*?)\K(([^,\s\/])\s*)\K>)|(<font\s?)\K>', 0) #find closing > after <font 
regexes[(3, (255, 128, 0))] = (r'(?i)(?<=<font)[^>]*?\Kface(?=[\s]*\=)', 0) #font modifiers
regexes[(4, (255, 128, 0))] = (r'(?i)(?<=<font)[^>]*?\Ksize(?=[\s]*\=)', 0) #font modifiers
regexes[(5, (255, 128, 0))] = (r'(?i)(?<=<font)[^>]*?\Kcolor(?=[\s]*\=)', 0) #font modifiers
regexes[(6, (255, 128, 0))] = (r'(?i)(?<=<font)[^>]*?\Kalpha(?=[\s]*\=)', 0) #font modifiers
regexes[(7, (230, 230, 0))] = (r'(?i)\[Access Rating(,\s?[0-1])?\]|\[AccurateRipResult(,\s?[0-1])?\]|\[Actors(,\s?[0-1])?\]|\[Album Artist \(auto\)(,\s?[0-1])?\]|\[Album Artist(,\s?[0-1])?\]|\[Album Popularity(,\s?[0-1])?\]|\[Album Type(,\s?[0-1])?\]|\[Album(,\s?[0-1])?\]|\[Altitude(,\s?[0-1])?\]|\[Amazon ASIN(,\s?[0-1])?\]|\[AMG Album ID(,\s?[0-1])?\]|\[AMG Album Pick(,\s?[0-1])?\]|\[AMG Moods(,\s?[0-1])?\]|\[AMG Rating(,\s?[0-1])?\]|\[AMG Styles(,\s?[0-1])?\]|\[AMG Track Genres(,\s?[0-1])?\]|\[AMG Track Pick(,\s?[0-1])?\]|\[AMG Track Styles(,\s?[0-1])?\]|\[Aperture(,\s?[0-1])?\]|\[Artist - Album \(Year\)(,\s?[0-1])?\]|\[Artist - Album - Composition(,\s?[0-1])?\]|\[Artist - Composition(,\s?[0-1])?\]|\[Artist ID(,\s?[0-1])?\]|\[Artist Popularity(,\s?[0-1])?\]|\[Artist(,\s?[0-1])?\]|\[Aspect Ratio(,\s?[0-1])?\]|\[Audio Analysis Info(,\s?[0-1])?\]|\[Author(,\s?[0-1])?\]|\[AutoEQ(,\s?[0-1])?\]|\[Available Assets(,\s?[0-1])?\]|\[Awards-BestActor(,\s?[0-1])?\]|\[Awards-BestActress(,\s?[0-1])?\]|\[Awards-BestDirector(,\s?[0-1])?\]|\[Awards-BestPicture(,\s?[0-1])?\]|\[Awards-Cinematography(,\s?[0-1])?\]|\[Awards-OscarsWon(,\s?[0-1])?\]|\[Awards-Roster(,\s?[0-1])?\]|\[Awards(,\s?[0-1])?\]|\[Band(,\s?[0-1])?\]|\[Bios(,\s?[0-1])?\]|\[Bit Depth(,\s?[0-1])?\]|\[Bitrate(,\s?[0-1])?\]|\[Bookmark(,\s?[0-1])?\]|\[BPM(,\s?[0-1])?\]|\[Budget(,\s?[0-1])?\]|\[Camera(,\s?[0-1])?\]|\[Caption(,\s?[0-1])?\]|\[Catalog #(,\s?[0-1])?\]|\[Category Image Category(,\s?[0-1])?\]|\[Channels(,\s?[0-1])?\]|\[Cinematographer(,\s?[0-1])?\]|\[Comment(,\s?[0-1])?\]|\[Complete Album(,\s?[0-1])?\]|\[Composer(,\s?[0-1])?\]|\[CompositionExclude(,\s?[0-1])?\]|\[CompositionGroup(,\s?[0-1])?\]|\[CompositionOrder(,\s?[0-1])?\]|\[Composition(,\s?[0-1])?\]|\[Compression(,\s?[0-1])?\]|\[Conductor(,\s?[0-1])?\]|\[Copyright(,\s?[0-1])?\]|\[Country(,\s?[0-1])?\]|\[Credits(,\s?[0-1])?\]|\[Critic Rating(,\s?[0-1])?\]|\[Custom(,\s?[0-1])?\]|\[Date \(day\)(,\s?[0-1])?\]|\[Date \(filename friendly\)(,\s?[0-1])?\]|\[Date \(month\)(,\s?[0-1])?\]|\[Date \(release\)(,\s?[0-1])?\]|\[Date \(Year\)(,\s?[0-1])?\]|\[Date Acquired(,\s?[0-1])?\]|\[Date Created(,\s?[0-1])?\]|\[Date First Rated(,\s?[0-1])?\]|\[Date Imported(,\s?[0-1])?\]|\[Date Last Opened(,\s?[0-1])?\]|\[Date Modified(,\s?[0-1])?\]|\[Date Recorded(,\s?[0-1])?\]|\[Date(,\s?[0-1])?\]|\[Day(,\s?[0-1])?\]|\[Description(,\s?[0-1])?\]|\[Dimensions(,\s?[0-1])?\]|\[Direction(,\s?[0-1])?\]|\[Director(,\s?[0-1])?\]|\[Disc #(,\s?[0-1])?\]|\[Discogs_Rating(,\s?[0-1])?\]|\[Discogs_Release_ID(,\s?[0-1])?\]|\[Disk Drive(,\s?[0-1])?\]|\[Distributor(,\s?[0-1])?\]|\[DLNA Server Info(,\s?[0-1])?\]|\[DRC(,\s?[0-1])?\]|\[DSP(,\s?[0-1])?\]|\[Duration(,\s?[0-1])?\]|\[DVD Video Info(,\s?[0-1])?\]|\[Dynamic Range \(DR\)(,\s?[0-1])?\]|\[Dynamic Range \(R128\)(,\s?[0-1])?\]|\[Edit Info(,\s?[0-1])?\]|\[Editor(,\s?[0-1])?\]|\[Encoding Settings(,\s?[0-1])?\]|\[Episode Air Sequence(,\s?[0-1])?\]|\[Episode URL(,\s?[0-1])?\]|\[Episode(,\s?[0-1])?\]|\[EQ(,\s?[0-1])?\]|\[Events(,\s?[0-1])?\]|\[Feed URL(,\s?[0-1])?\]|\[File Size(,\s?[0-1])?\]|\[File Type(,\s?[0-1])?\]|\[Filename \(name\)(,\s?[0-1])?\]|\[Filename \(path\)(,\s?[0-1])?\]|\[Filename(,\s?[0-1])?\]|\[Film Genre(,\s?[0-1])?\]|\[Flash(,\s?[0-1])?\]|\[Flickr Info(,\s?[0-1])?\]|\[Focal Length(,\s?[0-1])?\]|\[FPS-JRiver_Orig(,\s?[0-1])?\]|\[FPS-mediainfo(,\s?[0-1])?\]|\[FPS-mediainfo_Orig_FrameRate(,\s?[0-1])?\]|\[FPS-pics_gspot(,\s?[0-1])?\]|\[FPS(,\s?[0-1])?\]|\[Genre ID(,\s?[0-1])?\]|\[Genres-Autometa(,\s?[0-1])?\]|\[Genre(,\s?[0-1])?\]|\[Get Cover Art Info(,\s?[0-1])?\]|\[Get Metadata Info(,\s?[0-1])?\]|\[Gross Revenue(,\s?[0-1])?\]|\[GroupingDisplay(,\s?[0-1])?\]|\[Grouping(,\s?[0-1])?\]|\[GroupSet(,\s?[0-1])?\]|\[HDCD(,\s?[0-1])?\]|\[Height(,\s?[0-1])?\]|\[Image File(,\s?[0-1])?\]|\[Image Height(,\s?[0-1])?\]|\[Image Width(,\s?[0-1])?\]|\[IMDb ID(,\s?[0-1])?\]|\[IMDb Rating(,\s?[0-1])?\]|\[IMDb Votes(,\s?[0-1])?\]|\[Initial Key(,\s?[0-1])?\]|\[Instrument(,\s?[0-1])?\]|\[ISO(,\s?[0-1])?\]|\[Keywords(,\s?[0-1])?\]|\[Language(,\s?[0-1])?\]|\[Last Played(,\s?[0-1])?\]|\[Last Skipped(,\s?[0-1])?\]|\[Latitude(,\s?[0-1])?\]|\[Length In PCM Blocks(,\s?[0-1])?\]|\[Lens(,\s?[0-1])?\]|\[Library Server Info(,\s?[0-1])?\]|\[Library Server Key(,\s?[0-1])?\]|\[Library Server URL(,\s?[0-1])?\]|\[Link Start(,\s?[0-1])?\]|\[Link(,\s?[0-1])?\]|\[Longitude(,\s?[0-1])?\]|\[Lyricist(,\s?[0-1])?\]|\[Lyrics(,\s?[0-1])?\]|\[Media Sub Type(,\s?[0-1])?\]|\[Media Type(,\s?[0-1])?\]|\[Metascore(,\s?[0-1])?\]|\[Mix Album(,\s?[0-1])?\]|\[Month(,\s?[0-1])?\]|\[Mood(,\s?[0-1])?\]|\[Movement Count(,\s?[0-1])?\]|\[Movement Number(,\s?[0-1])?\]|\[Movement(,\s?[0-1])?\]|\[MPAA Rating Description(,\s?[0-1])?\]|\[MPAA Rating(,\s?[0-1])?\]|\[Music By(,\s?[0-1])?\]|\[NameSorting(,\s?[0-1])?\]|\[Name(,\s?[0-1])?\]|\[New Release(,\s?[0-1])?\]|\[Notes \(Album\)(,\s?[0-1])?\]|\[Notes(,\s?[0-1])?\]|\[Number Of Tracks(,\s?[0-1])?\]|\[Number Plays(,\s?[0-1])?\]|\[Orchestra(,\s?[0-1])?\]|\[Orientation(,\s?[0-1])?\]|\[Original Title(,\s?[0-1])?\]|\[Peak Level \(R128\)(,\s?[0-1])?\]|\[Peak Level \(Sample\)(,\s?[0-1])?\]|\[People(,\s?[0-1])?\]|\[Period(,\s?[0-1])?\]|\[Picture Format(,\s?[0-1])?\]|\[Picture Quality(,\s?[0-1])?\]|\[Picture Shape(,\s?[0-1])?\]|\[Places(,\s?[0-1])?\]|\[Playable(,\s?[0-1])?\]|\[Playback Channel Key(,\s?[0-1])?\]|\[Playback Device(,\s?[0-1])?\]|\[Playback Info(,\s?[0-1])?\]|\[Playback Range(,\s?[0-1])?\]|\[Popularity(,\s?[0-1])?\]|\[Pre-Show(,\s?[0-1])?\]|\[Preference(,\s?[0-1])?\]|\[Prequel(,\s?[0-1])?\]|\[Producer(,\s?[0-1])?\]|\[Program ID(,\s?[0-1])?\]|\[Protection Type(,\s?[0-1])?\]|\[Publisher(,\s?[0-1])?\]|\[Purchase Date(,\s?[0-1])?\]|\[Ranking(,\s?[0-1])?\]|\[Rating IMDb(,\s?[0-1])?\]|\[Rating(,\s?[0-1])?\]|\[Read Only(,\s?[0-1])?\]|\[Recording Rule ID(,\s?[0-1])?\]|\[Released(,\s?[0-1])?\]|\[Removable(,\s?[0-1])?\]|\[Repeated Frames(,\s?[0-1])?\]|\[Rerun(,\s?[0-1])?\]|\[Rotation(,\s?[0-1])?\]|\[Rotten Tomatoes(,\s?[0-1])?\]|\[Runtime(,\s?[0-1])?\]|\[Sample Rate(,\s?[0-1])?\]|\[Screenwriter(,\s?[0-1])?\]|\[Season and Best(,\s?[0-1])?\]|\[Season(,\s?[0-1])?\]|\[Sequel(,\s?[0-1])?\]|\[Series Description(,\s?[0-1])?\]|\[Series ID(,\s?[0-1])?\]|\[Series(,\s?[0-1])?\]|\[Show Work Movement(,\s?[0-1])?\]|\[Shutter Speed(,\s?[0-1])?\]|\[Single Album(,\s?[0-1])?\]|\[Situation(,\s?[0-1])?\]|\[Skip Count(,\s?[0-1])?\]|\[Soloists(,\s?[0-1])?\]|\[Soundtrack(,\s?[0-1])?\]|\[Source(,\s?[0-1])?\]|\[Special Handling(,\s?[0-1])?\]|\[Stack Conversion Type(,\s?[0-1])?\]|\[Stack Files(,\s?[0-1])?\]|\[Stack Tag(,\s?[0-1])?\]|\[Stack Top(,\s?[0-1])?\]|\[Stack View(,\s?[0-1])?\]|\[Store Field: Album ID(,\s?[0-1])?\]|\[Store Field: Album Popularity(,\s?[0-1])?\]|\[Store Field: Amazon ASIN(,\s?[0-1])?\]|\[Store Field: Artist ID(,\s?[0-1])?\]|\[Store Field: Artist Popularity(,\s?[0-1])?\]|\[Store Field: Available Assets(,\s?[0-1])?\]|\[Store Field: Genre ID(,\s?[0-1])?\]|\[Store Field: New Release(,\s?[0-1])?\]|\[Store Field: Number Of Tracks(,\s?[0-1])?\]|\[Store Field: Ranking(,\s?[0-1])?\]|\[Store Field: Store Name(,\s?[0-1])?\]|\[Store Field: Track ID(,\s?[0-1])?\]|\[Store Name(,\s?[0-1])?\]|\[Studios(,\s?[0-1])?\]|\[Style(,\s?[0-1])?\]|\[Sync(,\s?[0-1])?\]|\[Tag Line(,\s?[0-1])?\]|\[Tempo(,\s?[0-1])?\]|\[Text(,\s?[0-1])?\]|\[TheTVDB Series ID(,\s?[0-1])?\]|\[Thumbnail Info \(Local\)(,\s?[0-1])?\]|\[Thumbnail Info(,\s?[0-1])?\]|\[Thumbnail Offset Large(,\s?[0-1])?\]|\[Thumbnail Offset Medium(,\s?[0-1])?\]|\[Thumbnail Offset Small(,\s?[0-1])?\]|\[Thumbnail Position(,\s?[0-1])?\]|\[TMDb ID(,\s?[0-1])?\]|\[TMDb Score(,\s?[0-1])?\]|\[TMdbRating(,\s?[0-1])?\]|\[Total Discs(,\s?[0-1])?\]|\[Total Tracks(,\s?[0-1])?\]|\[Track #(,\s?[0-1])?\]|\[Track ID(,\s?[0-1])?\]|\[TV Channel Number(,\s?[0-1])?\]|\[TV Channel(,\s?[0-1])?\]|\[TV Tuner Input Type(,\s?[0-1])?\]|\[TV XMLTV ID(,\s?[0-1])?\]|\[UPC(,\s?[0-1])?\]|\[Usage Reporting Info(,\s?[0-1])?\]|\[Use Bookmarking(,\s?[0-1])?\]|\[User(,\s?[0-1])?\]|\[Video Codec(,\s?[0-1])?\]|\[Video DisplayAspect(,\s?[0-1])?\]|\[Video Format(,\s?[0-1])?\]|\[Video ScanType(,\s?[0-1])?\]|\[Video Standard(,\s?[0-1])?\]|\[Volume Adjustment(,\s?[0-1])?\]|\[Volume Level \(R128\)(,\s?[0-1])?\]|\[Volume Level \(ReplayGain\)(,\s?[0-1])?\]|\[Volume Name(,\s?[0-1])?\]|\[Waveform(,\s?[0-1])?\]|\[Web Media Info(,\s?[0-1])?\]|\[Web Media Search(,\s?[0-1])?\]|\[Web Media URL(,\s?[0-1])?\]|\[Website(,\s?[0-1])?\]|\[Width(,\s?[0-1])?\]|\[Work(,\s?[0-1])?\]|\[Year - Album(,\s?[0-1])?\]|\[Year(,\s?[0-1])?\]|\[Zone Last Opened(,\s?[0-1])?\]', 0)


# Definition of which area should not be styled
# 1 = comment style
# 2 = comment line style
# 16 = delimiter1
# ...
# 23 = delimiter8
#excluded_styles = [1, 2, 16, 17, 18, 19, 20, 21, 22, 23]
excluded_styles = [ ]
# ------------------------------------------------ /configuration area ---------------------------------------------------

try:
    EnhanceUDLLexer().main()
except NameError:

    user32 = wintypes.WinDLL('user32')

    WM_USER = 1024
    NPPMSG = WM_USER+1000
    NPPM_GETLANGUAGEDESC = NPPMSG+84
    SC_INDICVALUEBIT = 0x1000000
    SC_INDICFLAG_VALUEFORE = 1


    class SingletonEnhanceUDLLexer(type):
        '''
            Ensures, more or less, that only one
            instance of the main class can be instantiated
        '''
        _instance = None
        def __call__(cls, *args, **kwargs):
            if cls._instance is None:
                cls._instance = super(SingletonEnhanceUDLLexer, cls).__call__(*args, **kwargs)
            return cls._instance


    class EnhanceUDLLexer(object):
        '''
            Provides additional color options and should be used in conjunction with the built-in UDL function.
            An indicator is used to avoid style collisions.
            Although the Scintilla documentation states that indicators 0-7 are reserved for the lexers,
            indicator 0 is used because UDL uses none internally.

            Even when using more than one regex, it is not necessary to define more than one indicator
            because the class uses the flag SC_INDICFLAG_VALUEFORE.
            See https://www.scintilla.org/ScintillaDoc.html#Indicators for more information on that topic
        '''
        __metaclass__ = SingletonEnhanceUDLLexer

        def __init__(self):
            '''
                Instantiated the class,
                because of __metaclass__ = ... usage, is called once only.
            '''
            editor.callbackSync(self.on_updateui, [SCINTILLANOTIFICATION.UPDATEUI])
            notepad.callback(self.on_langchanged, [NOTIFICATION.LANGCHANGED])
            notepad.callback(self.on_bufferactivated, [NOTIFICATION.BUFFERACTIVATED])
            self.doc_is_of_interest = False
            self.lexer_name = None
            self.npp_hwnd = user32.FindWindowW(u'Notepad++', None)
            self.configure()


        @staticmethod
        def rgb(r, g, b):
            '''
                Helper function
                Retrieves rgb color triple and converts it
                into its integer representation

                Args:
                    r = integer, red color value in range of 0-255
                    g = integer, green color value in range of 0-255
                    b = integer, blue color value in range of 0-255
                Returns:
                    integer
            '''
            return (b << 16) + (g << 8) + r


        @staticmethod
        def paint_it(color, pos, length):
            '''
                This is where the actual coloring takes place.
                Color, the position of the first character and
                the length of the text to be colored must be provided.
                Coloring occurs only if the position is not within the excluded range.

                Args:
                    color = integer, expected in range of 0-16777215
                    pos = integer,  denotes the start position
                    length = integer, denotes how many chars need to be colored.
                Returns:
                    None
            '''
            if pos < 0 or editor.getStyleAt(pos) in excluded_styles:
                return
            editor.setIndicatorCurrent(0)
            editor.setIndicatorValue(color)
            editor.indicatorFillRange(pos, length)


        def style(self):
            '''
                Calculates the text area to be searched for in the current document.
                Calls up the regexes to find the position and
                calculates the length of the text to be colored.
                Deletes the old indicators before setting new ones.

                Args:
                    None
                Returns:
                    None
            '''
            start_line = editor.docLineFromVisible(editor.getFirstVisibleLine())
            end_line = editor.docLineFromVisible(start_line + editor.linesOnScreen())
            start_position = editor.positionFromLine(start_line)
            end_position = editor.getLineEndPosition(end_line)
            editor.setIndicatorCurrent(0)
            editor.indicatorClearRange(0, editor.getTextLength())
            for color, regex in self.regexes.items():
                editor.research(regex[0],
                                lambda m: self.paint_it(color[1],
                                                        m.span(regex[1])[0],
                                                        m.span(regex[1])[1] - m.span(regex[1])[0]),
                                0,
                                start_position,
                                end_position)


        def configure(self):
            '''
                Define basic indicator settings, the needed regexes as well as the lexer name.

                Args:
                    None
                Returns:
                    None
            '''
            editor1.indicSetStyle(0, INDICATORSTYLE.TEXTFORE)
            editor1.indicSetFlags(0, SC_INDICFLAG_VALUEFORE)
            editor2.indicSetStyle(0, INDICATORSTYLE.TEXTFORE)
            editor2.indicSetFlags(0, SC_INDICFLAG_VALUEFORE)
            self.regexes = OrderedDict([ ((k[0], self.rgb(*k[1]) | SC_INDICVALUEBIT), v) for k, v in regexes.items() ])
            self.lexer_name = u'User Defined language file - %s' % lexer_name


        def check_lexer(self):
            '''
                Checks if the current document is of interest
                and sets the flag accordingly

                Args:
                    None
                Returns:
                    None
            '''
            language = notepad.getLangType()
            length = user32.SendMessageW(self.npp_hwnd, NPPM_GETLANGUAGEDESC, language, None)
            buffer = ctypes.create_unicode_buffer(u' ' * length)
            user32.SendMessageW(self.npp_hwnd, NPPM_GETLANGUAGEDESC, language, ctypes.byref(buffer))
            self.doc_is_of_interest = True if buffer.value == self.lexer_name or 'JRiver' in buffer.value else False
		

        def on_bufferactivated(self, args):
            '''
                Callback which gets called every time one switches a document.
                Triggers the check if the document is of interest.

                Args:
                    provided by notepad object but none are of interest
                Returns:
                    None
            '''
            self.check_lexer()


        def on_updateui(self, args):
            '''
                Callback which gets called every time scintilla
                (aka the editor) changed something within the document.

                Triggers the styling function if the document is of interest.

                Args:
                    provided by scintilla but none are of interest
                Returns:
                    None
            '''
            if self.doc_is_of_interest:
                self.style()


        def on_langchanged(self, args):
            '''
                Callback gets called every time one uses the Language menu to set a lexer
                Triggers the check if the document is of interest

                Args:
                    provided by notepad object but none are of interest
                Returns:
                    None
            '''
            self.check_lexer()


        def main(self):
            '''
                Main function entry point.
                Simulates two events to enforce detection of current document
                and potential styling.

                Args:
                    None
                Returns:
                    None
            '''
            self.on_bufferactivated(None)
            self.on_updateui(None)

    EnhanceUDLLexer().main()


